//   Snafu.java --- SMA de recherche d'information sur l'Internet

//   Copyright (C) 2000 Valvassori Mose <djedi@altern.org>

//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.

//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package Snafu;
import java.net.*;
import java.io.*;
import java.util.*;


/**
   Server d'Agent.<br>
   C'est sur cette classe que vont venir se connecter les 
   <em>agents Client</em>

   @author Valvassori Mose
   @version alpha
 */
public class SnafuAgentServer extends SnafuAgent {
  public SnafuAgentServer(){
    super("SnafuAgentServer","server");
    ability_list = new Vector();
    commande_en_cours = new Hashtable();
  }






  /**
     Cette fonction est la premire  tre appel dans la main.
     Elle met en place une liaison sur le port 17023 et cre un 
     nouveau canal des qu'on se connecte dessus.

     @author Valvassori Mose
   */
  protected synchronized void init(){
    try {
      sSocket = new ServerSocket(port);
    } catch (IOException e) {
      System.out.println("On ne peut pas couter le port "+port);
      System.exit(-1);
    }

    //    System.out.println ("Je suis pass par la "+getID()); // DEBUG

    servert = new SASThread(sSocket,this);
    servert.start();    
  }



  /**
     Reoit les messages et les traitent.
   */
  protected void run(){
    String messg;
    SnafuReponse repon;
    SASCommande commande;

    while (true){
      messg=message.recevoir();
      System.out.println (""+message.getReceiver());
      if ((message.getReceiver().equals("null")) || 
	  (message.getReceiver().equals(getID())) ||
	  (message.getReceiver().equals(getAbility()))){

	//	System.out.println ("Content:"+messg);	// DEBUG
	if (message.getAction().equals("reply")){
	  repon = new SnafuReponse(messg);
	  //	  System.out.println ("Nom: "+repon.getNom()); // DEBUG
	  System.out.println ("Reply: "+message.getIn_reply_to()); // DEBUG
	  System.out.println ("URL: "+repon.getAdresse()); // DEBUG
	  //	  System.out.println ("Desc: "+repon.getDescription());	// DEBUG
	  commande =(SASCommande)commande_en_cours.get(message.getIn_reply_to());
	  message.init_entries();
	  message.setContent(messg);
	  message.setIn_reply_to(commande.getReply_to());
	  message.setReceiver(commande.getSender());
	  message.setPerformative("reply");
	  message.envoyer();
	}
	else if (message.getAction().equals("tell")){
	  if (messg.equals("(end:)")){
	    System.out.println ("Perte d'un canal");
	  } else if (messg.startsWith("(ability:")){
	    if (!ability_list.contains(messg.substring(10,messg.length()-1)))
	      ability_list.addElement(messg.substring(10,messg.length()-1));
	  }
	}
	else if (message.getAction().equals("achieve")){
	  System.out.println ("***"+message.getSender());	// DEBUG
	  commande = new SASCommande(message.getSender(),messg,message.getReply_to());
	  commande_en_cours.put(message.getReply_to(),commande);
	  message.init_entries();
	  message.setContent(messg);
	  message.setReply_to(commande.getReply_to());
	  message.setReceiver("search");
	  message.setPerformative("achieve");
	  message.envoyer();
	}
      }
	message.amorce_reception();	// Dit  KQML que l'on est pret  recevoir un nouveau message
    } // Fin While
  }










  // ***** Main *****
  public static void main(String[] args){
    SnafuAgentServer server = new SnafuAgentServer();
    System.out.println ("Server.");
    server.init();
    server.run();

    server = null;
    System.exit(0); 
  }







  /**
     Destructeur de SnafuAgentServer.<br>
     Arrete le SASThread.
   */
  protected void finalize() throws Throwable {
    servert.stop();
    servert = null;
    System.out.println ("Destruction Serveur");	// DEBUG
    super.finalize();
  }



  // --------------------
  private ServerSocket sSocket;
  private static final int port=17023;
  private SASThread servert;
  private Vector ability_list;
  private Hashtable commande_en_cours;






  //          ****************************************
  //********************* Classes internes *********************
  //          ****************************************

  /**
     Cette classe sert  collecter les connections entrantes. Elle
     coute le server de socket et ajoute un nouveau canal lorsque 
     il le faut...

     @author Valvassori Mose
     @version alpha
   */
  private class SASThread extends Thread {
    SASThread(ServerSocket s,SnafuAgentServer serv){
      super("SASThread");
      sSocket=s;
      server=serv;
    }


    public void run(){
      while (true){
	try{
	  server.ajoute_canal(sSocket.accept());
	  try{
	    sleep(10);
	  } catch (InterruptedException e) {
	    System.out.println (""+e);
	  }
	} catch (IOException e){
	  System.out.println ("Erreur:"+e);
	}
      }
    }
    private ServerSocket sSocket;
    private SnafuAgentServer server;
  } // Fin de SASThread



  /**
     Cette classe sert  stocker les commandes en cours qui
     sont traites par le serveur.
     @author Valvassori Mose
     @version alpha
   */
  private class SASCommande {
    SASCommande (String se,String co,String re){
      sender=se;
      content=co;
      reply_to=re;
    }
    public String getContent(){
      return (content);
    }
    public String getSender(){
      return (sender);
    }
    public String getReply_to(){
      return (reply_to);
    }
    public String toString(){
      return (new String("s:"+sender+"\nc:"+content));
    }
    private String sender;
    private String content;
    private String reply_to;
  } // fin SASCommande
}



