//   SnafuKQML.java --- SMA de recherche d'information sur l'Internet

//   Copyright (C) 2000 Benjamin Drieu <bdrieu@april.org>

//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.

//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


package Snafu;
import gnu.regexp.*;
import java.lang.*;
import java.util.*;

/**
 *  SnafuKQML permet de construire une requte KQML, et de la dcoder
 *  au besoin pour en construire une chane qui peut alors tre
 *  balance sur un canal quelconque.
 *
 * @author Benjamin Drieu <bdrieu@april.org>, Valvassori Mose <mvalvassori@april.org>
 * @version alpha
 * 
 */
public class SnafuKQML extends Thread
{

  private String action;

  private String content;
  private String reply_to;
  private String in_reply_to;
  private String receiver;
  private String sender; 
  private String sender1; 
  private String language="snafu";
  private SnafuAgent agent;
  
  /**
   * Spcifie l'expression rgulire qui matche l'action passe via
   * une chane KQML.
   */
  private static final String begining = "^\\(([a-z]+)";
  private static final String couples = ":([a-zA-Z-]+)[ \n]+(\\([^\\)]+\\)|[^ \n\\(\\)]+)";

  /**
     Reprsente l'ensemble des connections avec les autres agents.
     @see SnafuKQML()
     @see #recoit_canal
     @see SnafuAgentCanal
  */
  private Vector canaux;
  private int canal_pret;
  private boolean canaux_vide;

  /**
   * Constructeur vide, a peut servir.
   * 
   */
  public SnafuKQML() 
  {
    canaux = new Vector();
  }


  /**
     Constructeur initialisant l'emteur du message.
     @param Sender l'mteur du message
   */
 public SnafuKQML(String Sender,String Abil,SnafuAgent a){
    canaux=new Vector();
    sender1=Sender;
    agent = a;
  }
  


  public String getAction(){
    return action;
  }
  public String getContent(){
    return content;
  }
  public String getReply_to(){
    return reply_to;
  }
  public String getIn_reply_to(){
    return in_reply_to;
  }
  public String getReceiver(){
    return receiver;
  }
  public String getSender(){
    return sender;
  }
  public void setContent(String s){
    content=s;
  }
  public void setReply_to(String s){
    reply_to=s;
  }
  public void setIn_reply_to(String s){
    in_reply_to=s;
  }
  public void setPerformative(String s){
    action=s;
  }
  public void setReceiver(String s){
    receiver=s;
  }
  /**
     Dit si des messages sont en attente de dpart ou de reception.
     @return <tt>true</tt>des message sont en attente <tt>false</tt> sinon.
     @see #canaux_vide
   */
  public boolean message_attente(){
    return (!canaux_vide);
  }

  /**
     Attend de recevoir un message. Lorsque un message est reu,
     on previent.
   */
  public void run(){

    System.out.println ("Dmarage SnafuKQML");		// DEBUG
    while (true){
      canaux_vide=true;
      for (int i=0; i<get_nb_canaux(); i++){
	if (((SnafuAgentCanal)canaux.elementAt(i)).pret()){
	  canaux_vide=false;
	  a_recu(i);
	}
      }

      try{
     	sleep(100);
      } catch (InterruptedException e) {
     	System.out.println ("SnafuKQML IE (run)"+e);
      } 
    }
  }


  /** Trigger qui reamorce la reception d'un message aprs l'avoir
      lu
  */
  public synchronized void amorce_reception(){
    try{
      notifyAll();
    } catch (IllegalMonitorStateException e){
      //	    System.out.println ("Notify pas bon: "+e);
    }     
  }


  /** 
      Debloque la mthode recevoir et attend d'tre
      reamorcer.
      @see #amorce_reception
      @see #recevoir()
  */
  private synchronized void a_recu(int i){
    notify();
    canal_pret = i;
    try{
      wait();				// Attend le reamorcage
    } catch (InterruptedException e){
      System.out.println ("KQML run "+e);
    }
  }


  /**
   * Cette fonction produit une version texte (conversion en object
   * String) de l'object SnafuKQML (ce qui permet par exemple de
   * l'imprimer).
   */
  public String toString()
  {
    return "(" + action +
      " :sender " + sender1 +
      " :content " + content +
      " :language " + language +
      " :receiver " + receiver +
      " :reply-to " + reply_to +
      " :in-reply-to " + in_reply_to +
      ")";
  }
    

  /**
   * Cette fonction prend en argument une chane de caractres qui
   * est la requte KQML  dcoder, et la dcode.  Elle utilise les
   * expressions rgulires spcifies par la variable d'instance
   * <i>begining</i> et <i>couples</i> pour extraire l'action
   * spcifie par la requte KQML et les couples nom/valeur
   * attachs  l'action.  Cette fonction se charge ensuite de
   * remplir les variables d'instance sender, reply_to, ... selon le
   * contenu de la requte KQML.  Peut-tre vaudrait-il mieux
   * utiliser une table de hash ?
   * @param string La chane  dcoder.
   * @exception REException Ncessaire pour utiliser l'object RE.
   * @exception SnafuKQMLException La requte KQML n'est pas conforme.
   */
  public void decode (String string) throws  SnafuKQMLException
  {

    if (string != null)
      // Rcupration de l'action
      try{
	RE re = new RE(begining);
	REMatch match = re.getMatch(string);
	if (match == null) 
	  {
	    throw new SnafuKQMLException(string+" didn't match "+begining);
	  }
	action = match.toString(1);
	
	// Rcupration des couples  :nom valeur 
	re = new RE(couples);
	REMatch [] matches = re.getAllMatches(string, match.getEndIndex());
	for (int i = 0; i < matches.length; i++)
	  {
	    if (matches[i].toString(1).equals("sender"))
	      {
		sender = matches[i].toString(2);
	      }
	    else if (matches[i].toString(1).equals("reply-to"))
	      {
		reply_to = matches[i].toString(2);
	      }
	    else if (matches[i].toString(1).equals("in-reply-to"))
	      {
		in_reply_to = matches[i].toString(2);
	      }
	    else if (matches[i].toString(1).equals("receiver"))
	      {
		receiver = matches[i].toString(2);
	      }
	    else if (matches[i].toString(1).equals("content"))
	      {
		content = matches[i].toString(2);
	      }
	    else if (matches[i].toString(1).equals("language"))
	      {
		language = matches[i].toString(2);
	      }
	  }
      } catch (REException e){
      }
  }



  /**
     Ajouter un nouveau canal de communication. Le canal est aussi activ
     @param c le nouveau canal de communication
     @see canaux
  */
  protected void recoit_canal (SnafuAgentCanal c,String Abil){
    c.start();
    canaux.addElement(c);
    presentation(Abil);
  } 


  /**
     Envoie une chaine sur tous les canaux.
  */
  public void envoyer(){
    if (!canaux.isEmpty())
      for (int i=0; i<get_nb_canaux(); i++)
	try{
	  ((SnafuAgentCanal)canaux.elementAt(i)).envoyer(this.toString());
	} catch (ArrayIndexOutOfBoundsException e){	
	}
  }

  /**
     Recoit uin message dans un canal. Cette mthode attend de recevoir
     un message.
     @return renvoie le contenu du message ou null si il y a erreur.
  */
  public synchronized String recevoir(){
    //    if (!canaux.isEmpty()){
    try {
      wait();
    } catch (InterruptedException e) {
      System.out.println ("SnafuKQML erreur Interruption: "+e);
    }
    
    //    System.out.println ("*** A recu un truc"); // DEBUG

    return recevoir(canal_pret);
    //  } else
    //    return null;
  }
  /**
     Envoie une chaine quelconque sur un canal.
  */
  protected void envoyer(int i,String s){
    if (!canaux.isEmpty())
      try{
	((SnafuAgentCanal)canaux.elementAt(i)).envoyer(s);
      } catch (ArrayIndexOutOfBoundsException e){	
      }
  }

  /**
     Reoit une chane d'un canal donne.
  */
  protected String recevoir(int i){
    if (!canaux.isEmpty())
      try{
	if (((SnafuAgentCanal)canaux.elementAt(i)).pret() == true){
	  decode(((SnafuAgentCanal)canaux.elementAt(i)).recevoir());
	  return content;
	}
	else
	  return null;
      } catch (ArrayIndexOutOfBoundsException e){
	return null;
      } catch (SnafuKQMLException e){
	System.out.println (""+e);
	return null;
      }
    else
      return null;
  }


  /**
     Cette fonction efface les champs qui ne servent pas.
  */
  public void init_entries(){
    action=null;
    content=null;
    reply_to=null;
    in_reply_to=null;
    receiver=null;  
  }



  /**
     Renvoie le nombre de canaux.
     @return le nombre de canaux
     @see SnafuAgentCanal
   */
  public int get_nb_canaux(){
    return canaux.size();
  }



  /**
     Destructeur de la classe KQML.
     Avant de disparatre, elle envoie un faire-part de dcs.
   */
  protected void finalize() throws SnafuKQMLException {
    System.out.println ("Destruction de KQML");	// DEBUG
    init_entries();
    action="tell";
    content="(end:)";
    envoyer();
    try{
      super.finalize();
    } catch(Throwable e){
      throw new SnafuKQMLException(e.getMessage());
    } 
  }



  /**
     Envoie un message de presentation a un collegue agent.
  */
  private void presentation(String Abil){
    init_entries();
    action="tell";
    content="(ability: "+Abil+")";
    envoyer();
  }
}



/**
 * Cette classe sert  lever les exceptions de la classe SnafuKQML.
 * @author Benjamin Drieu <bdrieu@april.org>
 * @version alpha
 */
class SnafuKQMLException extends Exception
{
  public SnafuKQMLException () {}
  public SnafuKQMLException (String msg) 
  {
    super(msg);
  }
}


   ////////////////////////////////////////////////////////////////////////////////////////////////
   //  /**										         //
   //   * Constructeur mettant  jour les variables de l'instance en			         //
   //   * fonction d'une chane passe en argument.  Cette chane est			         //
   //   * encode via la mthode encode.						         //
   //   * @param string La chane de caractre  dcoder.				         //
   //   * @see #decode									         //
   //   * @exception REException Ncessaire pour utiliser l'object RE.			         //
   //   */										         //
   //  public SnafuKQML(String string) throws gnu.regexp.REException, SnafuKQMLException         //
   //  {										         //
   //  	this.decode(string);								         //
   //  	canaux = new Vector();								         //
   //  }										         //
   ////////////////////////////////////////////////////////////////////////////////////////////////
 
